#include <malloc.h>
#include <gccore.h>
#include "globals.h"
#include "sprite.h"



static void copy_from_framebuffer (u32 *fb_ptr, u32 *dest, u16 width,
				   u16 height, u16 x_padding)
{
  u32 x, y ;
  for (y = 0; y < height; y++) {
    for (x = 0; x < (width >> 1); x++) {
      *dest = *fb_ptr ;
      dest++ ;
      fb_ptr++ ;
    }
    /* Go to next row. */
    fb_ptr = fb_ptr + (x_padding >> 1) ;
  }
}



static void copy_to_framebuffer (u32 *src, u32 *fb_ptr, u16 width, u16 height,
				 u16 x_padding)
{
  u32 x, y ;
  for (y = 0; y < height; y++) {
    for (x = 0; x < (width >> 1); x++) {
      *fb_ptr = *src ;
      src++ ;
      fb_ptr++ ;
    }
    /* Row done, go to next one. */
    fb_ptr = fb_ptr + (x_padding >> 1) ;
  }
}



static void mask_copy_to_framebuffer (u32 *src, u8 *mask, u32 *fb_ptr,
				      u16 width, u16 height, u16 x_padding)
{
  u32 x, y ;
  u32 pixel ;
  u32 bg ;

  for (y = 0; y < height; y++) {
    for (x = 0; x < (width >> 1); x++) {
      pixel = 0 ;
      bg = *fb_ptr ;
      if ((*mask) == 0x00) pixel |= (bg & 0xFFFF0000) ;
      else pixel |= ((*src) & 0xFFFF0000) ;
      mask++ ;
      if ((*mask) == 0x00) pixel |= (bg & 0x0000FFFF) ;
      else pixel |= ((*src) & 0x0000FFFF) ;
      *fb_ptr = pixel ;
      mask++;
      src++ ;
      fb_ptr++ ;
    }
    /* Row done, go to next one. */
    fb_ptr = fb_ptr + (x_padding >> 1) ;
  }
}



struct sprite *alloc_sprite (u16 width, u16 height, u32 *pixels, u8 *mask)
{
  struct sprite *sprite ;

  /* Sprites MUST be of even width. */
  if ((width & 0x01) != 0) return (NULL) ;

  sprite = (struct sprite*) malloc (sizeof (struct sprite)) ;
  if (sprite == NULL) return (NULL) ;

  /* Since we directly work on u32, we need half bytes on a row. */
  sprite->saved_bg = (u32*) malloc (((width >> 1) * height) * sizeof (u32)) ;
  if (sprite->saved_bg == NULL) {
    free (sprite) ;     /* Free the sprite's structure. */
    return (NULL) ;
  }

  sprite->top_left_x = 0 ;
  sprite->top_left_y = 0 ;
  sprite->width = width ;
  sprite->height = height ;
  sprite->pixels = pixels ;
  sprite->mask = mask ;
  sprite->displayed = 0 ;    /* Sprites are not displayed by default. */
  return (sprite) ;
}



void free_sprite (struct sprite *sprite)
{
  free (sprite->saved_bg) ;
  free (sprite) ;
}



void show_sprite (struct sprite *sprite)
{
  if (! sprite->displayed) {
    u16 x_padding ;  /* How many columns to skip to the beginning of the next
			row in the framebuffer. */
    u32 *fb_ptr ;

    x_padding = rmode->fbWidth - sprite->width ;
    /* First, save the background under the sprite's position. */
    fb_ptr =
      &frameBuffer[sprite->top_left_y * (rmode->fbWidth >> 1) +
		   (sprite->top_left_x >> 1)] ;
    copy_from_framebuffer
      (fb_ptr, sprite->saved_bg, sprite->width, sprite->height, x_padding) ;

    /* Now copy the sprite pixels to the framebuffer at the new position. */
    if (sprite->mask == NULL)
      copy_to_framebuffer
	(sprite->pixels, fb_ptr, sprite->width, sprite->height, x_padding) ;
    else
      mask_copy_to_framebuffer
	(sprite->pixels, sprite->mask, fb_ptr, sprite->width, sprite->height,
	 x_padding) ;

    sprite->displayed = 1 ;
  }
}



void hide_sprite (struct sprite *sprite)
{
  if (sprite->displayed) {
    u16 x_padding ;  /* How many columns to skip to the beginning of the next
			row in the framebuffer. */
    u32 *fb_ptr ;

    x_padding = rmode->fbWidth - sprite->width ;
    /* First, we must restore the saved background. */
    fb_ptr =
      &frameBuffer[sprite->top_left_y * (rmode->fbWidth >> 1) +
		   (sprite->top_left_x >> 1)] ;
    copy_to_framebuffer
      (sprite->saved_bg, fb_ptr, sprite->width, sprite->height, x_padding) ;
    /* Make the sprite non visible. */
    sprite->displayed = 0 ;
  }
}



void move_sprite (struct sprite *sprite, s16 x, s16 y)
{
  u32 *fb_ptr ;
  u16 x_padding ;  /* How many columns to skip to the beginning of the next
		      row in the framebuffer. */

  /* For the moment, simply clip by ignoring the position if the sprite is
     partly outside the screen area. */

  if ((x < 0) || (x > (rmode->fbWidth - sprite->width))) return ;
  if ((y < 0) || (y > (rmode->xfbHeight - sprite->height))) return ;

  /* If the sprite is not visible, we don't need to manage the pixels copy and
     backup. */
  if (sprite->displayed) {
    x_padding = rmode->fbWidth - sprite->width ;

    /* First, we must restore the saved background. */
    fb_ptr =
      &frameBuffer[sprite->top_left_y * (rmode->fbWidth >> 1) +
		   (sprite->top_left_x >> 1)] ;
    copy_to_framebuffer
      (sprite->saved_bg, fb_ptr, sprite->width, sprite->height, x_padding) ;

    /* Now, save the background under the new position. */
    fb_ptr = &frameBuffer[y * (rmode->fbWidth >> 1) + (x >> 1)] ;
    copy_from_framebuffer
      (fb_ptr, sprite->saved_bg, sprite->width, sprite->height, x_padding) ;

    /* Now copy the sprite pixels to the framebuffer at the new position. */
    if (sprite->mask == NULL)
      copy_to_framebuffer
	(sprite->pixels, fb_ptr, sprite->width, sprite->height, x_padding) ;
    else
      mask_copy_to_framebuffer
	(sprite->pixels, sprite->mask, fb_ptr, sprite->width, sprite->height,
	 x_padding) ;
  }

  /* Always update the sprite's position. */
  sprite->top_left_x = x ;
  sprite->top_left_y = y ;
}
